/*----------------------------------------------------------------------*
 * AdjList.c
 *
 * Michael Geikowski, 22.3.93
 *
 * Enthaelt die Member-Funktionen der Klassen Node und NodeList.
 *----------------------------------------------------------------------*/
/* Olaf Meier, 16.6.94
 *
 * Ergaenzungen:
 * -------------
 * Zusaetzliche Member- und NonMemberroutinen zur Behandlung von Parsetree-Knoten und -Kanten
 * sowie die 3 neuen privaten Variablen PTminpos, PTmaxpos (Positionsangaben des
 * jeweiligen Parsetree-Unterbaumes bzgl. des Spezifikationsmoduls) und PTpredecessor
 * (Vorgaenger des aktuellen Knotens - NULL, falls Wurzelknoten.)
 *-----------------------------------------------------------------------------*/ 


#include <assert.h>
#include <string.h>

#include "adjlistgnu.h"

/*----------------------------------------------------------------------*
 * HilfsKlasse Edge
 *----------------------------------------------------------------------*/

class Edge : public ListElem
//==========================
	// Ein Objekt der Klasse Edge beschreibt eine Ausgangskante eines
        // Knotens durch Angabe des Zielknotens und eines Kantentyps. Objekte
        // der Klasse Edge werden zu den Adjazenzlisten der Knoten eines
        // Graphen zusammengefasst.
 {
 public:

 friend class Node;

   Edge(Node *n,char t)
 //--------------------
 	: DestNode(n), EdgeType(t)
	{}
   ~Edge()
 //-------
	{}

 private:

   Node *DestNode;	// Zielknoten der Kante
   char EdgeType;	// Typ der Kante
 };


/*---------------------------------------------------------------------*
 * Hilfsklasse ActionElem
 *---------------------------------------------------------------------*/

class ActionElem : public ListElem
//================================
  // Speichert alle Aktionen, in denen auf die betreffende Systemvariable 
  // zugegriffen wird.

{ 
 public :
  
  friend class Node;

  ActionElem(int acnr = 0)  // Aktionsnummer 0 ist eine dummy Aktion
   : actionNr(acnr)
  {}

  ~ActionElem()
  {}

 
  int GetActionNr()

  { return actionNr;}

 private :

  int actionNr;

};
   
  
 
/*----------------------------------------------------------------------*
 * Klasse Node
 *----------------------------------------------------------------------*/

Node::Node(char* name)
 {
 assert(name);
 nodeName = new char[strlen(name)+1];	// Kopiere Namen
 strcpy(nodeName,name);
 PTminpos = 0;
 PTmaxpos = 0;
 NodeType = 0;				// Initialisiere Zaehler
 Level = 0;
 NrOfInEdges = NrOfOutEdges = 0;
 Pos = TempPos = 0;
 Counter = Counter2 = 0;
 OutEdges = new List;
 PTpredecessor = NULL;
 ActionList = NULL; 

 }

Node::~Node()
 {
 if (nodeName) delete nodeName;
 if (OutEdges) delete OutEdges;
 if (ActionList) delete ActionList;  
 }

void Node::AddEdge(Node *n, char EdgeType)
 {
 assert(n && OutEdges);
 OutEdges->NewFirstElem(new Edge(n,EdgeType));
 NrOfOutEdges++;
 n->NrOfInEdges++;
 }


void Node::PTAddEdge(Node* n, char EdgeType)
 {
  assert(n && OutEdges);
  n->SetPTpredecessor(this); // Vorgaengerkoten von 'n' ist 'this'
  OutEdges->Append(new Edge(n,EdgeType));

  NrOfOutEdges++;
  n->NrOfInEdges++;
 }

Node* Node::GetFirstEdge()
 {
 assert (OutEdges);
 Edge *e = (Edge*) OutEdges->GetFirstElem();
 if (e)
   {
   return e->DestNode;
   }
 else
   {
   return NULL;
   }
 }
 
Node* Node::GetNextEdge()
 {
 assert (OutEdges);
 Edge *e = (Edge*) OutEdges->GetNextElem();
 if (e)
   {
   return e->DestNode;
   }
 else
   {
   return NULL;
   }
 }

int Node::GetFirstActionNr()
 { 
 assert(ActionList);
 ActionElem*  ae = (ActionElem*) ActionList->GetFirstElem();
 if(ae)
  {
   return ae->GetActionNr();
  }
 else
  {
   return -1;  // Liste ist leer
  }
 }

int Node::GetNextActionNr()
 {
 assert(ActionList);
 ActionElem* ae = (ActionElem*) ActionList->GetNextElem();
 if(ae)
  {
   return ae->GetActionNr();
  }
 else
  {
   return -1;  // Ende der Liste erreicht
  }
 }   

void Node::AddActionNr(int actionnr)
 {
  ActionElem* ae = (ActionElem*) ActionList->GetFirstElem();
  while(ae)
   { 
      if(ae->GetActionNr() == actionnr)    // Aktion ist schon eingetragen
       return;
      else     
       ae = (ActionElem*) ActionList->GetNextElem(); 
   }
  ae = new ActionElem(actionnr);   // Aktion wird neu hinzugefuegt
  ActionList->Append(ae);
 }
     

void Node::ChangeCurrentEdge(Node* n)
 {
 assert(n && OutEdges);
 Edge* e = (Edge*) OutEdges->GetCurrentElem();
 assert(e);
 Node* m = e->DestNode;
 assert(m);
 m->NrOfInEdges--;
 e->DestNode = n;
 n->NrOfInEdges++;
 }

char Node::GetTypeOfCurrentEdge()
 {
 assert(OutEdges);
 Edge* e = (Edge*) OutEdges->GetCurrentElem();
 assert(e);
 return e->EdgeType;
 }

Node* Node::PTCopyNode(int level)
 { assert(level >= 1);
   Node *source,*dest;
   char t = GetNodeType();
   char* nodename = GetName();
   long min = GetPTminpos();
   long max = GetPTmaxpos();
   
   source = new Node(nodename);
   source->SetPTminpos(min);
   source->SetPTmaxpos(max);

   dest = GetFirstEdge();
   if((!dest) || (level = 1))
    {
      source->SetNodeType('D');  // D = 'Downwards
      return source;
    }
   else
    { source->SetNodeType(t);
      while(dest)
       {
         t = GetTypeOfCurrentEdge();
         source->PTAddEdge(dest->PTCopyNode(level - 1), t);
         dest = GetNextEdge();
       };
      return source;
     }
  }  

int Node::HasPTpredecessor()
   {
     return (PTpredecessor != NULL);
   } 

int Node::HasPTsuccessor()
   {
     Node* n = GetFirstEdge();
     return (n != NULL);
   }

Node* Node::PTGoUpwards(int level)
 { assert(level >= 1);
   Node* pred = GetPTpredecessor();   // Vorgaenger von 'this'
   if(pred && (level > 1)) // ...weitere Vorgaenger sind vorhanden
    { return (pred->PTGoUpwards(level -1));}
   else                    //... weitere Vorgaenger sind nicht vorhanden !
    { return this;}
 }

/*----------------------------------------------------------------------*
 * Klasse NodeList
 *----------------------------------------------------------------------*/

Node* NodeList::AddNode(char* name)
 {
 assert(name);
 Node *n = (Node*) GetFirstElem();	// Pruefe, ob Knoten mit diesem Namen
 while (n)				// schon bekannt.
   {
   if (strcmp(name,n->GetName())==0)
     {
     return n;				// Bekannt => alten Knoten zurueckgeben
     }
   n = (Node*) GetNextElem();
   }
 NewFirstElem(new Node(name));		// Unbekannt => Neuen Knoten erzeugen
 return (Node*) GetFirstElem();
 }


Node* NodeList::PTAddNode(char* name, char nt, long min, long max)
 {
 assert(name);
 Node* nod;
 Node *n = (Node*) GetFirstElem();     // Pruefe, ob Knoten mit diesem Namen und den Subtreepositionen
 while(n)                              // min und max schon bekannt.
  {
   if ( 
        (strcmp(name, n->GetName()) == 0) &&
        (n->GetPTminpos() == min) &&
        (n->GetPTmaxpos() == max) )
       { 
         return n;                      // Bekannt => alten Knoten zurueckgeben
       }
    n = (Node*) GetNextElem();
  } 
  nod = new Node(name);                 // Unbekannt => Neuen Knoten erzeugen
  nod->SetNodeType(nt);
  nod->SetPTminpos(min);
  nod->SetPTmaxpos(max);
 
  Append(nod);
  
  return (Node*) GetLastElem();
 }


Node* NodeList::PTSearchforNode(Node* n)
  { assert(n); 
    char* nodename = n->GetName();
    long min = n->GetPTminpos();
    long max = n->GetPTmaxpos();
    Node* node = GetFirstNode();
    while(node)   // Uberpruefung, ob Knoten in Liste enthalten ist
                 
     {
       if( (node->GetPTminpos() == min) &&        
           (node->GetPTmaxpos() == max) &&
           (strcmp(nodename, node->GetName()) == 0) ) 
            { break; }  // gefunden !
       node = GetNextNode();
      }
    return node;
  }   


Node* NodeList::PTSearchforNodeLevel(int l)
 { assert(l >= 0);
   Node* n;
   n = GetFirstNode();
   while(n)
    {
      if(n->GetLevel() == l)
       { break;}
      n = GetNextNode();
    }
   return n;
  }

Node* NodeList::SCAddNode(char* name, int actionnr,char nt)
 {  assert(name);
    ActionElem* ael;
    Node *n = (Node*) GetFirstElem();      // Pruefe, ob Knoten mit diesem Namen
    while (n)                              // schon bekannt.
     {
      if (strcmp(name,n->GetName())==0)   // Knoten ist bereits vorhanden
       {  
          ael = (ActionElem*) n->ActionList->GetFirstElem();
          while(ael)
           {   
               if(ael->GetActionNr() == actionnr)  // Aktionsnummer ist schon vorhanden
                return n;
               ael = (ActionElem*) n->ActionList->GetNextElem();
           }
          ael = new ActionElem(actionnr);  // Aktionsnummer wird hinzugefuegt.
          n->ActionList->Append(ael);
          return n;
       }
      n = (Node*) GetNextElem();
     }
    n = new Node(name);     // Knoten und Aktionsmnummer sind noch nicht vorhanden.
    n->SetNodeType(nt);
    n->ActionList = new List;
    ael = new ActionElem(actionnr);
    n->ActionList->Append(ael);
    Append(n); 
    return (Node*) GetLastElem();               
 }


/*------------------------------------------------------------------------*
 * Nonmember - Routinen
 *------------------------------------------------------------------------*/

void AddNodetoAdjlist(Node* n, AdjList* l)
 { assert(n && l);
   Node* next;
   l->Append(n);
   next = n->GetFirstEdge();
   while(next)
    {
      AddNodetoAdjlist(next, l);
      next = n->GetNextEdge();
    };
 }   


void SCAddEdgesToNodes(AdjList* adjl)
 {
   assert(adjl);
   int i,j,mini, count,NrOfNodes = adjl->GetLength();
   Node *n,*help; 
   Node **NodeArray = new Node*[NrOfNodes];
   count = 0;
   n =  adjl->GetFirstNode();
   while( n && (count < NrOfNodes))
    {
      NodeArray[count] =  n;
      count++;
      n = adjl->GetNextNode();
    }

  // Knoten  werden nach aufsteigender Knotennamenlaenge sortiert

  for(i = 1; i < NrOfNodes-1; i++)
  {
   mini = i;
   for(j = i+1; j < NrOfNodes; j++)
     if( strlen(NodeArray[j]->GetName()) < strlen(NodeArray[mini]->GetName()) )
           mini = j;

   help = NodeArray[i], NodeArray[i] =  NodeArray[mini] , NodeArray[mini] = help;
  }
  // Aufbau des Systemvariablenbaums:
  // Vaterknoten wird mit entsprechenden Sohnen verbunden
  // (Teilkomponenten der entsprechenden Systemvariablen)

  for(i = NrOfNodes -1 ; i > 0; i--)
   {
     for( j = i - 1; j > 0; j--)
      {     
        if(strncmp(NodeArray[i]->GetName(),NodeArray[j]->GetName(),strlen(NodeArray[j]->GetName())) == 0 )
           {
             NodeArray[j]->AddEdge(NodeArray[i],'x');
             break;
           }
      }
   }   


  help = adjl->GetFirstNode();  // top Knoten '_STATE_'
  if(help)
   {
    n = adjl->GetNextNode();
    while(n) 
     { 
        if(n->GetNrOfInEdges() == 0)  // nicht korrekt benutzte Systemvariablen
                                      // (die nicht in der Systemvariablenliste auftauchen)
                                      // werden markiert. 
         {
           help->AddEdge(n,'E');
         }
        n = adjl->GetNextNode();
     }
   }

  delete [NrOfNodes] NodeArray;

    }

